@@ -27,11 +27,16 @@
 #include <string.h>
 #include <assert.h>
 #include <errno.h>
-#include <signal.h>
 #include <fcntl.h>
-#include <unistd.h>
-#include <sys/time.h>
-#include <sys/types.h>
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
+#ifdef HAVE_SYS_TIME_H
+# include <sys/time.h>
+#endif
+#ifdef HAVE_SYS_TYPES_H
+# include <sys/types.h>
+#endif
 #include <glib.h>
 #include <windows.h>
 #include <io.h>
@@ -111,7 +116,7 @@ static struct
 static GIOChannel *
 find_channel (int fd)
 {
-  if (fd < 0 || fd >= MAX_SLAFD)
+  if (fd < 0 || fd >= MAX_SLAFD || !giochannel_table[fd].used)
     return NULL;
 
   return giochannel_table[fd].chan;
